import abc
from typing import TypeVar

from oaas.client_definition import ClientDefinition

T = TypeVar("T")


class ClientMiddleware(metaclass=abc.ABCMeta):
    """
    Defines a serialization provider that can interpret the
    registrations and invoke the methods.
    """

    @abc.abstractmethod
    def create_client(self, client_definition: ClientDefinition) -> T:
        ...

    @abc.abstractmethod
    def can_handle(self, client_definition: ClientDefinition) -> bool:
        pass



from typing import Callable, TypeVar, Type, Optional, Dict

import oaas._registrations as registrations
from oaas.client_definition import ClientDefinitionMetadata, ClientDefinition
from oaas.client_provider import ClientMiddleware
from oaas.server_provider import ServerMiddleware
from oaas.service_definition import ServiceDefinition, ServiceDefinitionMetadata

T = TypeVar("T")


def client(
    name: str,
    namespace: str = "default",
    version: str = "1",
    tags: Dict[str, str] = None,
) -> Callable[..., Type[T]]:
    """
    Declare a service from the system. All the input and output data
    should be serializable. The serialization format depends on
    the provider being used. To get an instance to the client, call
    `get_client`.
    """

    def wrapper_builder(t: Type[T]) -> Type[T]:
        cd = ClientDefinition(
            name=name,
            namespace=namespace,
            version=version,
            code=t,
            tags=tags,
        )
        registrations.clients[t] = cd

        return t

    return wrapper_builder


def service(
    name: str,
    *,
    namespace: str = "default",
    version: str = "1",
    tags: Optional[Dict[str, str]] = None,
) -> Callable[..., Type[T]]:
    """
    Mark a service to be exposed to the system. All the input
    and output data should be serializable. The serialization
    format depends on the provider being used.
    """

    def wrapper_builder(t: Type[T]) -> Type[T]:
        sd = ServiceDefinition(
            namespace=namespace,
            name=name,
            version=version,
            code=t,
            tags=tags,
        )
        registrations.services.append(sd)

        return t

    return wrapper_builder


def serve() -> None:
    """
    Expose all the defined services using the underlying
    providers.
    """
    services_without_middleware = set()

    for service in registrations.services:
        if not _has_server_middleware(service):
            services_without_middleware.add(service)

    if services_without_middleware:
        raise Exception(
            "Some services have no backing middleware. Make sure the "
            "middleware servers are added using oaas.register_server_provider() "
            f"before calling serve(): {services_without_middleware}"
        )

    for provider in registrations.servers_middleware:
        provider.serve()


def _has_server_middleware(service) -> bool:
    for provider in registrations.servers_middleware:
        if provider.can_serve(service):
            return True

    return False


def join() -> None:
    """
    Wait for all the defined servers to come down.
    """
    for provider in registrations.servers_middleware:
        provider.join()


def get_client(t: Type[T]) -> T:
    """
    Create a client for the given type.
    """
    for provider in registrations.clients_middleware:
        client_definition = registrations.clients[t]
        if provider.can_handle(client_definition):
            return provider.create_client(client_definition)

    raise Exception(
        f"No serialization provider was registered to handle " f"{t} clients."
    )


def register_server_provider(server_middleware: ServerMiddleware):
    """
    Register a serialization provider. Normally this should be taken
    care by the middleware.
    """
    registrations.servers_middleware.add(server_middleware)


def register_client_provider(client_middleware: ClientMiddleware):
    """
    Register a serialization provider. Normally this should be taken
    care by the middleware.
    """
    registrations.clients_middleware.add(client_middleware)



from typing import Dict, Optional, Any, TypeVar, Type, Callable

T = TypeVar("T")
ClientDefinitionMetadata = Dict[str, Any]


class ClientDefinition:
    def __init__(
        self,
        *,
        namespace: str = "default",
        name: str,
        version: str = "1",
        code: Callable[..., T],
        tags: Optional[Dict[str, str]] = None,
    ) -> None:
        self.namespace = namespace
        self.name = name
        self.version = version
        self.code = code
        self.tags = tags

    @property
    def gav(self) -> str:
        return f"{self.namespace}:{self.name}:{self.version}"



import unittest
from typing import Any, Dict

import oaas
from test.local_serialization_provider import LocalClientServerMiddleware


@oaas.service("datastore")
class DataStoreService:
    def __init__(self) -> None:
        self._items: Dict[str, Any] = dict()

    def put_item(self, key: str, value: Any) -> None:
        self._items[key] = value

    def get_item(self, key: str) -> Any:
        if key in self._items:
            return self._items.get(key)

        return None

    def remove_item(self, key: str) -> None:
        del self._items[key]


@oaas.client("datastore")
class DataStore:
    def put_item(self, key: str, value: Any) -> None:
        ...

    def get_item(self, key: str) -> Any:
        ...

    def remove_item(self, key: str) -> None:
        ...


provider = LocalClientServerMiddleware()
oaas.register_server_provider(provider)
oaas.register_client_provider(provider)

oaas.serve()


class TestServiceRegistrar(unittest.TestCase):
    def test_service_calls(self) -> None:
        data_store = oaas.get_client(DataStore)

        self.assertIsNone(data_store.get_item("a"))
        data_store.put_item("a", "avalue")
        self.assertEqual("avalue", data_store.get_item("a"))
        data_store.remove_item("a")
        self.assertIsNone(data_store.get_item("a"))



import abc
from typing import TypedDict, List, Optional


class ServiceDefinitionRequired(TypedDict):
    name: str
    protocol: str


class ServiceDefinition(ServiceDefinitionRequired, total=False):
    namespace: str
    version: str


IpLink = str


class ServiceAddressRequired(TypedDict):
    port: int
    addresses: List[IpLink]


class ServiceAddress(ServiceAddressRequired, total=False):
    pass


class OaasRegistry(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def register_service(
        self, service: ServiceDefinition, address: ServiceAddress
    ) -> None:
        ...

    @abc.abstractmethod
    def resolve_service(self, service: ServiceDefinition) -> Optional[ServiceAddress]:
        ...



from collections import OrderedDict
from typing import Dict, Set, Type, List

from oaas.client_definition import ClientDefinition
from oaas.client_provider import ClientMiddleware
from oaas.server_provider import ServerMiddleware
from oaas.service_definition import ServiceDefinition

clients: Dict[Type, ClientDefinition] = OrderedDict()
services: List[ServiceDefinition] = []

servers_middleware: Set[ServerMiddleware] = set()
clients_middleware: Set[ClientMiddleware] = set()



import abc
from typing import TypeVar

from oaas.service_definition import ServiceDefinition

T = TypeVar("T")


class ServerMiddleware(metaclass=abc.ABCMeta):
    """
    Defines a serialization provider that can interpret the
    registrations and invoke the methods.
    """

    @abc.abstractmethod
    def serve(self) -> None:
        ...

    @abc.abstractmethod
    def join(self) -> None:
        ...

    @abc.abstractmethod
    def can_serve(self, service: ServiceDefinition) -> bool:
        ...



from typing import Any, Dict, cast, TypeVar

import oaas._registrations as registrations
from oaas import ServiceDefinition
from oaas.client_definition import ClientDefinition
from oaas.client_provider import ClientMiddleware
from oaas.server_provider import ServerMiddleware

T = TypeVar("T")


class ReflectionInvoker:
    def __init__(self, *, delegate: Any) -> None:
        self._delegate = delegate

    def __getattr__(self, item):
        return getattr(self._delegate, item)


class LocalClientServerMiddleware(ClientMiddleware, ServerMiddleware):
    """
    Provides a server that can invoke local services
    """

    def __init__(self) -> None:
        self._service_instance: Dict[str, Any] = dict()

    def serve(self) -> None:
        for service_definition in registrations.services:
            self._service_instance[service_definition.name] = service_definition.code()

    def join(self) -> None:
        pass

    def create_client(self, client_definition: ClientDefinition) -> T:
        service_instance = self._service_instance[client_definition.name]

        return cast(T, ReflectionInvoker(delegate=service_instance))

    def can_handle(self, client_definition: ClientDefinition) -> bool:
        return True

    def can_serve(self, service: ServiceDefinition) -> bool:
        return True



from setuptools import setup
from setuptools import find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

packages = find_packages()

setup(
    name="oaas",
    version="1.6.1",
    description="Operation as a Service",
    long_description=readme,
    author="Bogdan Mustiata",
    author_email="bogdan.mustiata@gmail.com",
    license="BSD",
    install_requires=[],
    packages=packages,
    package_data={
        "": ["*.txt", "*.rst"],
        "oaas": ["py.typed"],
    },
)
